import json
import math
import os
import warnings
from typing import Dict, List, Optional, Union

from jina.orchestrate.deployments.config.k8slib import kubernetes_tools
from jina.serve.networking import GrpcConnectionPool

PERIOD_SECONDS = 5


def get_template_yamls(
    name: str,
    namespace: str,
    image_name: str,
    container_cmd: str,
    container_args: str,
    replicas: int,
    pull_policy: str,
    jina_deployment_name: str,
    pod_type: str,
    shard_id: Optional[int] = None,
    env: Optional[Dict] = None,
    env_from_secret: Optional[Dict] = None,
    image_pull_secrets: Optional[List] = None,
    gpus: Optional[Union[int, str]] = None,
    image_name_uses_before: Optional[str] = None,
    image_name_uses_after: Optional[str] = None,
    container_cmd_uses_before: Optional[str] = None,
    container_cmd_uses_after: Optional[str] = None,
    container_args_uses_before: Optional[str] = None,
    container_args_uses_after: Optional[str] = None,
    monitoring: bool = False,
    protocol: Optional[Union[str, List[str]]] = None,
    volumes: Optional[List[str]] = None,
    timeout_ready: int = 600000,
) -> List[Dict]:
    """Get the yaml description of a service on Kubernetes

    :param name: name of the service and deployment
    :param namespace: k8s namespace of the service and deployment
    :param image_name: image for the k8s deployment
    :param container_cmd: command executed on the k8s pods
    :param container_args: arguments used for the k8s pod
    :param replicas: number of replicas
    :param pull_policy: pull policy used for fetching the Docker images from the registry.
    :param jina_deployment_name: Name of the Jina Deployment this deployment belongs to
    :param pod_type: type os this pod, can be gateway/head/worker
    :param shard_id: id of this shard, None if shards=1 or this is gateway/head
    :param env: environment variables to be passed into configmap.
    :param env_from_secret: environment variables from secret to be passed to this pod
    :param image_pull_secrets: list of secrets to be added uder ImagePullSecrets
    :param gpus: number of gpus to use, for k8s requires you pass an int number, refers to the number of requested gpus.
    :param image_name_uses_before: image for uses_before container in the k8s deployment
    :param image_name_uses_after: image for uses_after container in the k8s deployment
    :param container_cmd_uses_before: command executed in the uses_before container on the k8s pods
    :param container_cmd_uses_after: command executed in the uses_after container on the k8s pods
    :param container_args_uses_before: arguments used for uses_before container on the k8s pod
    :param container_args_uses_after: arguments used for uses_after container on the k8s pod
    :param monitoring: enable monitoring on the deployment
    :param protocol: In case of being a Gateway, the protocol or protocols list used to expose its server
    :param volumes: If volumes are passed to Executors, Jina will create a StatefulSet instead of Deployment and include the first volume in the volume mounts
    :param timeout_ready: The timeout in milliseconds of a Pod waits for the runtime to be ready. This parameter will be
        reflected in Kubernetes in the startup configuration where the failureThreshold will be calculated depending on
        timeout_ready. Value -1 is not supported for kubernetes
    :return: Return a dictionary with all the yaml configuration needed for a deployment
    """
    # we can always assume the ports are the same for all executors since they run on different k8s pods
    # port expose can be defined by the user
    port = (
        [GrpcConnectionPool.K8S_PORT + i for i in range(len(protocol))]
        if isinstance(protocol, list)
        else GrpcConnectionPool.K8S_PORT
    )  # TODO: This cannot happen
    port_monitoring = GrpcConnectionPool.K8S_PORT_MONITORING

    # we cast port to list of ports and protocol to list of protocols
    if not isinstance(port, list):
        ports = [port]
    else:
        ports = port
    if not isinstance(protocol, list):
        protocols = [protocol]
    else:
        protocols = protocol

    if timeout_ready == -1:
        warnings.warn(
            'timeout_ready=-1 is not supported, setting timeout_ready to 10 minutes'
        )
        timeout_ready = 600000
    failure_threshold = max(math.ceil((timeout_ready / 1000) / PERIOD_SECONDS), 3)

    template_params = {
        'name': name,
        'namespace': namespace,
        'image': image_name,
        'replicas': replicas,
        'command': container_cmd,
        'args': container_args,
        'port': ports[0],
        'port_uses_before': GrpcConnectionPool.K8S_PORT_USES_BEFORE,
        'port_uses_after': GrpcConnectionPool.K8S_PORT_USES_AFTER,
        'args_uses_before': container_args_uses_before,
        'args_uses_after': container_args_uses_after,
        'command_uses_before': container_cmd_uses_before,
        'command_uses_after': container_cmd_uses_after,
        'image_uses_before': image_name_uses_before,
        'image_uses_after': image_name_uses_after,
        'pull_policy': pull_policy,
        'jina_deployment_name': jina_deployment_name,
        'shard_id': f'\"{shard_id}\"' if shard_id is not None else '\"\"',
        'pod_type': pod_type,
        'env_from_secret': env_from_secret,
        'image_pull_secrets': image_pull_secrets,
        'protocol': str(protocols[0]).lower() if protocols[0] is not None else '',
        'volume_path': volumes[0] if volumes is not None else None,
        'period_seconds': PERIOD_SECONDS,
        'failure_threshold': failure_threshold,
    }

    if gpus and gpus != 'all':
        template_params['device_plugins'] = {'nvidia.com/gpu': gpus}

    template_name = 'deployment-executor' if name != 'gateway' else 'deployment-gateway'

    template_params['ports-section'] = ''.join(
        [f'\n            - containerPort: {_p}' for _p in ports]
    )

    if volumes:
        template_name = 'statefulset-executor'
        template_params['accessModes'] = json.loads(
            os.environ.get('JINA_K8S_ACCESS_MODES', '["ReadWriteOnce"]')
        )
        template_params['storageClassName'] = os.environ.get(
            'JINA_K8S_STORAGE_CLASS_NAME', 'standard'
        )
        template_params['storageCapacity'] = os.environ.get(
            'JINA_K8S_STORAGE_CAPACITY', '10G'
        )
    elif image_name_uses_before and image_name_uses_after:
        template_name = 'deployment-uses-before-after'
    elif image_name_uses_before:
        template_name = 'deployment-uses-before'
    elif image_name_uses_after:
        template_name = 'deployment-uses-after'

    if monitoring:
        service_yaml = kubernetes_tools.get_yaml(
            'service_monitoring',
            {
                'name': name,
                'target': name,
                'namespace': namespace,
                'port': ports[0],
                'type': 'ClusterIP',
                'port_monitoring': port_monitoring,
            },
        )

        service_monitor_yaml = kubernetes_tools.get_yaml(
            'service_monitor',
            {
                'name': name,
                'target': name,
                'namespace': namespace,
            },
        )
    else:
        service_yaml = kubernetes_tools.get_yaml(
            'service',
            {
                'name': name,
                'target': name,
                'namespace': namespace,
                'port': ports[0],
                'type': 'ClusterIP',
            },
        )
        service_monitor_yaml = None

    extra_services = [
        kubernetes_tools.get_yaml(
            'service',
            {
                'name': f'{name}-{i}-{protocol}'.lower(),
                'target': name,
                'namespace': namespace,
                'port': port,
                'type': 'ClusterIP',
            },
        )
        for i, (port, protocol) in enumerate(zip(ports[1:], protocols[1:]), start=1)
    ]

    template_yaml = kubernetes_tools.get_yaml(template_name, template_params)

    if 'JINA_LOG_LEVEL' in os.environ:
        env = env or {}
        env['JINA_LOG_LEVEL'] = os.environ['JINA_LOG_LEVEL']

    yamls = [
        kubernetes_tools.get_yaml(
            'configmap',
            {
                'name': name,
                'namespace': namespace,
                'data': env,
            },
        ),
        service_yaml,
        *extra_services,
        template_yaml,
    ]

    if service_monitor_yaml:
        yamls.append(service_monitor_yaml)

    return yamls



from jina.serve.runtimes.gateway.gateway import BaseGateway
from jina.serve.runtimes.servers.composite import CompositeServer

__all__ = ['CompositeGateway']


class CompositeGateway(CompositeServer, BaseGateway):
    """
    :class:`CompositeGateway` is a CompositeServer that can be loaded from YAML as any other Gateway
    """

    pass



import os
from typing import Dict

DEPLOYMENT_FILES = [
    'statefulset-executor',
    'deployment-executor',
    'deployment-gateway',
    'deployment-uses-before',
    'deployment-uses-after',
    'deployment-uses-before-after',
]

cur_dir = os.path.dirname(__file__)
DEFAULT_RESOURCE_DIR = os.path.join(
    cur_dir, '..', '..', '..', '..', 'resources', 'k8s', 'template'
)


def get_yaml(template: str, params: Dict) -> Dict:
    """Create a resource on Kubernetes based on the `template`. It fills the `template` using the `params`.

    :param template: path to the template file.
    :param params: dictionary for replacing the placeholders (keys) with the actual values.
    :return: The yaml dictionary with the corresponding template filled with parameters
    """
    if template == 'configmap':
        yaml = _get_configmap_yaml(template, params)
    elif template in DEPLOYMENT_FILES:
        yaml = _get_yaml(template, params)
        if params.get('device_plugins'):
            yaml = _get_deployment_with_device_plugins(yaml, params)
        if params.get('env_from_secret'):
            yaml = _get_deployment_with_env_secret(yaml, params)
        if params.get('image_pull_secrets'):
            yaml = _get_deployment_with_image_pull_secrets(yaml, params)
    else:
        yaml = _get_yaml(template, params)

    return yaml


def _get_yaml(template: str, params: Dict) -> Dict:
    import yaml

    path = os.path.join(DEFAULT_RESOURCE_DIR, f'{template}.yml')

    with open(path, encoding='utf-8') as f:
        content = f.read()
        for k, v in params.items():
            content = content.replace(f'{{{k}}}', str(v))
        d = yaml.safe_load(content)

    return d


def _get_configmap_yaml(template: str, params: Dict):
    import yaml

    path = os.path.join(DEFAULT_RESOURCE_DIR, f'{template}.yml')

    with open(path, encoding='utf-8') as f:
        config_map = yaml.safe_load(f)

    config_map['metadata']['name'] = params.get('name') + '-' + 'configmap'
    config_map['metadata']['namespace'] = params.get('namespace')
    if params.get('data'):
        for key, value in params['data'].items():
            config_map['data'][key] = str(value)
    return config_map


def _get_device_plugins(params: Dict):
    data = {'limits': {}}
    for key, value in params.items():
        data['limits'][key] = value
    return data


def _get_deployment_with_device_plugins(deployment: Dict, params: Dict) -> Dict:
    device_plugins = _get_device_plugins(params['device_plugins'])

    deployment['spec']['template']['spec']['containers'][0][
        'resources'
    ] = device_plugins
    return deployment


def _get_deployment_with_env_secret(deployment: Dict, params: Dict) -> Dict:
    for k, v in params['env_from_secret'].items():
        env_var = {
            'name': k,
            'valueFrom': {'secretKeyRef': {'name': v['name'], 'key': v['key']}},
        }

        deployment['spec']['template']['spec']['containers'][0]['env'].append(env_var)

    return deployment


def _get_deployment_with_image_pull_secrets(deployment: Dict, params: Dict) -> Dict:
    image_pull_secrets = params['image_pull_secrets']
    image_pull_secrets_dict = [{'name': secret} for secret in image_pull_secrets]
    deployment['spec']['template']['spec']['imagePullSecrets'] = image_pull_secrets_dict
    return deployment



import asyncio
import copy
from asyncio import Event, Task
from typing import Callable, Dict, List, Optional, TYPE_CHECKING, Union
from jina._docarray import docarray_v2

if not docarray_v2:
    from docarray import DocumentArray
else:
    from docarray import DocList
from jina.types.request.data import DataRequest

if TYPE_CHECKING:
    from jina._docarray import DocumentArray


class BatchQueue:
    """A batch queue that holds the data request and the callable to batch requests to."""

    def __init__(
            self,
            func: Callable,
            request_docarray_cls,
            response_docarray_cls,
            output_array_type: Optional[str] = None,
            params: Optional[Dict] = None,
            flush_all: bool = False,
            preferred_batch_size: int = 4,
            timeout: int = 10_000,
            custom_metric: Optional[Callable[['DocumentArray'], Union[int, float]]] = None,
            use_custom_metric: bool = False,
    ) -> None:
        # To keep old user behavior, we use data lock when flush_all is true and no allow_concurrent
        self.func = func
        if params is None:
            params = dict()
        self._is_closed = False
        self._output_array_type = output_array_type
        self.params = params
        self._request_docarray_cls = request_docarray_cls
        self._response_docarray_cls = response_docarray_cls
        self._flush_all = flush_all
        self._preferred_batch_size: int = preferred_batch_size
        self._custom_metric = None if not use_custom_metric else custom_metric
        self._metric_value = 0
        self._timeout: int = timeout
        self._reset()
        self._flush_trigger: Event = Event()
        self._timer_started, self._timer_finished = False, False
        self._timer_task: Optional[Task] = None

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(preferred_batch_size={self._preferred_batch_size}, timeout={self._timeout})'

    def __str__(self) -> str:
        return self.__repr__()

    def _reset(self) -> None:
        """Set all events and reset the batch queue."""
        self._requests: List[DataRequest] = []
        # a list of every request idx inside self._requests
        self._request_idxs: List[int] = []
        self._request_lens: List[int] = []
        self._docs_metrics: List[int] = []
        self._requests_completed: List[asyncio.Queue] = []
        if not docarray_v2:
            self._big_doc: DocumentArray = DocumentArray.empty()
        else:
            self._big_doc = self._request_docarray_cls()
        self._metric_value = 0

        self._flush_task: Optional[Task] = None
        self._flush_trigger: Event = Event()

    def _cancel_timer_if_pending(self):
        if (
                self._timer_task
                and not self._timer_task.done()
                and not self._timer_task.cancelled()
        ):
            self._timer_finished = False
            self._timer_task.cancel()

    def _start_timer(self):
        self._cancel_timer_if_pending()
        self._timer_task = asyncio.create_task(self._sleep_then_set())

    async def _sleep_then_set(self):
        """Sleep and then set the event"""
        self._timer_finished = False
        await asyncio.sleep(self._timeout / 1000)
        self._flush_trigger.set()
        self._timer_finished = True

    async def push(self, request: DataRequest, http=False) -> asyncio.Queue:
        """Append request to the the list of requests to be processed.

        This method creates an asyncio Queue for that request and keeps track of it. It returns
        this queue to the caller so that the caller can now when this request has been processed

        :param request: The request to append to the queue.
        :param http: Flag to determine if the request is served via HTTP for some optims

        :return: The queue that will receive when the request is processed.
        """
        docs = request.docs

        if not self._timer_task or self._timer_finished:
            # If there is no timer (first arrival), or the timer is already consumed, any new push should trigger a new Timer, before
            # this push requests the data lock. The order of accessing the data lock guarantees that this request will be put in the `big_doc`
            # before the `flush` task processes it.
            self._start_timer()
        if not self._flush_task:
            self._flush_task = asyncio.create_task(self._await_then_flush(http))
        self._big_doc.extend(docs)
        next_req_idx = len(self._requests)
        num_docs = len(docs)
        metric_value = num_docs
        if self._custom_metric is not None:
            metrics = [self._custom_metric(doc) for doc in docs]
            metric_value += sum(metrics)
            self._docs_metrics.extend(metrics)
        self._metric_value += metric_value
        self._request_idxs.extend([next_req_idx] * num_docs)
        self._request_lens.append(num_docs)
        self._requests.append(request)
        queue = asyncio.Queue()
        self._requests_completed.append(queue)
        if self._metric_value >= self._preferred_batch_size:
            self._flush_trigger.set()

        return queue

    async def _await_then_flush(self, http=False) -> None:
        """Process all requests in the queue once flush_trigger event is set.
        :param http: Flag to determine if the request is served via HTTP for some optims
        """

        def _get_docs_groups_completed_request_indexes(
                non_assigned_docs,
                non_assigned_docs_reqs_idx,
                sum_from_previous_mini_batch_in_first_req_idx,
                requests_lens_in_batch,
        ):
            """
            This method groups all the `non_assigned_docs` into groups of docs according to the `req_idx` they belong to.
            They are only distributed when we are sure that the request is full.

            :param non_assigned_docs: The documents that have already been processed but have not been assigned to a request result
            :param non_assigned_docs_reqs_idx: The request IDX that are not yet completed (not all of its docs have been processed)
            :param sum_from_previous_mini_batch_in_first_req_idx: The number of docs from previous iteration that belong to the first non_assigned_req_idx. This is useful to make sure we know when a request is completed.
            :param requests_lens_in_batch: List of lens of documents for each request in the batch.

            :return: list of document groups and a list of request Idx to which each of these groups belong
            """
            distributed_requests = []
            completed_req_idx = []
            num_distributed_docs = 0
            num_docs_in_req_idx = 0
            min_involved_req_idx = non_assigned_docs_reqs_idx[0]
            req_idx = min_involved_req_idx
            for req_idx in non_assigned_docs_reqs_idx:
                sum_from_previous_mini_batch_in_first_req_idx -= (
                    1  # the previous leftovers are being allocated here
                )
                if req_idx > min_involved_req_idx:
                    request_bucket = non_assigned_docs[
                                     num_distributed_docs: num_distributed_docs
                                                           + num_docs_in_req_idx
                                     ]
                    num_distributed_docs += num_docs_in_req_idx
                    completed_req_idx.append(min_involved_req_idx)
                    min_involved_req_idx = req_idx
                    num_docs_in_req_idx = 0
                    distributed_requests.append(request_bucket)
                num_docs_in_req_idx += 1

            if (
                    req_idx not in completed_req_idx
                    and num_docs_in_req_idx + sum_from_previous_mini_batch_in_first_req_idx
                    == requests_lens_in_batch[req_idx]
            ):
                completed_req_idx.append(req_idx)
                request_bucket = non_assigned_docs[
                                 num_distributed_docs: num_distributed_docs + num_docs_in_req_idx
                                 ]
                distributed_requests.append(request_bucket)

            return distributed_requests, completed_req_idx

        async def _assign_results(
                non_assigned_docs,
                non_assigned_docs_reqs_idx,
                sum_from_previous_mini_batch_in_first_req_idx,
                requests_lens_in_batch,
                requests_in_batch,
                requests_completed_in_batch,
        ):
            """
            This method aims to assign to the corresponding request objects the resulting documents from the mini batches.
            They are assigned when we are sure that the Request is fully processed.
            It also communicates to the corresponding queue that the request is full so that it can be returned

            :param non_assigned_docs: The documents that have already been processed but have not been assigned to a request result
            :param non_assigned_docs_reqs_idx: The request IDX that are not yet completed (not all of its docs have been processed)
            :param sum_from_previous_mini_batch_in_first_req_idx: The number of docs from previous iteration that belong to the first non_assigned_req_idx. This is useful to make sure we know when a request is completed.
            :param requests_lens_in_batch: List of lens of documents for each request in the batch.
            :param requests_in_batch: List requests in batch
            :param requests_completed_in_batch: List of queues for requests to be completed

            :return: amount of assigned documents so that some documents can come back in the next iteration
            """
            (
                docs_grouped,
                completed_req_idxs,
            ) = _get_docs_groups_completed_request_indexes(
                non_assigned_docs,
                non_assigned_docs_reqs_idx,
                sum_from_previous_mini_batch_in_first_req_idx,
                requests_lens_in_batch
            )
            num_assigned_docs = sum(len(group) for group in docs_grouped)

            for docs_group, request_idx in zip(docs_grouped, completed_req_idxs):
                request = requests_in_batch[request_idx]
                request_completed = requests_completed_in_batch[request_idx]
                if http is False or self._output_array_type is not None:
                    request.direct_docs = None  # batch queue will work in place, therefore result will need to read from data.
                    request.data.set_docs_convert_arrays(
                        docs_group, ndarray_type=self._output_array_type
                    )
                else:
                    request.direct_docs = docs_group
                await request_completed.put(None)

            return num_assigned_docs

        def batch(iterable_1, iterable_2, n: Optional[int] = 1, iterable_metrics: Optional = None):
            if n is None:
                yield iterable_1, iterable_2
                return
            elif iterable_metrics is None:
                items = len(iterable_1)
                for ndx in range(0, items, n):
                    yield iterable_1[ndx: min(ndx + n, items)], iterable_2[
                                                                ndx: min(ndx + n, items)
                                                                ]
            else:
                batch_idx = 0
                batch_weight = 0

                for i, (item, weight) in enumerate(zip(iterable_1, iterable_metrics)):
                    batch_weight += weight

                    if batch_weight >= n:
                        yield iterable_1[batch_idx: i + 1], iterable_2[batch_idx: i + 1]
                        batch_idx = i + 1
                        batch_weight = 0

                # Yield any remaining items
                if batch_weight > 0:
                    yield iterable_1[batch_idx: len(iterable_1)], iterable_2[batch_idx: len(iterable_1)]

        await self._flush_trigger.wait()
        # writes to shared data between tasks need to be mutually exclusive
        big_doc_in_batch = copy.copy(self._big_doc)
        requests_idxs_in_batch = copy.copy(self._request_idxs)
        requests_lens_in_batch = copy.copy(self._request_lens)
        docs_metrics_in_batch = copy.copy(self._docs_metrics)
        requests_in_batch = copy.copy(self._requests)
        requests_completed_in_batch = copy.copy(self._requests_completed)

        self._reset()

        # At this moment, we have documents concatenated in big_doc_in_batch corresponding to requests in
        # requests_idxs_in_batch with its lengths stored in requests_lens_in_batch. For each requests, there is a queue to
        # communicate that the request has been processed properly.

        if not docarray_v2:
            non_assigned_to_response_docs: DocumentArray = DocumentArray.empty()
        else:
            non_assigned_to_response_docs = self._response_docarray_cls()

        non_assigned_to_response_request_idxs = []
        sum_from_previous_first_req_idx = 0
        for docs_inner_batch, req_idxs in batch(
                big_doc_in_batch, requests_idxs_in_batch,
                self._preferred_batch_size if not self._flush_all else None, docs_metrics_in_batch if self._custom_metric is not None else None
        ):
            involved_requests_min_indx = req_idxs[0]
            involved_requests_max_indx = req_idxs[-1]
            input_len_before_call: int = len(docs_inner_batch)
            batch_res_docs = None
            try:
                batch_res_docs = await self.func(
                    docs=docs_inner_batch,
                    parameters=self.params,
                    docs_matrix=None,  # joining manually with batch queue is not supported right now
                    tracing_context=None,
                )
                # Output validation
                if (docarray_v2 and isinstance(batch_res_docs, DocList)) or (
                        not docarray_v2
                        and isinstance(batch_res_docs, DocumentArray)
                ):
                    if not len(batch_res_docs) == input_len_before_call:
                        raise ValueError(
                            f'Dynamic Batching requires input size to equal output size. Expected output size {input_len_before_call}, but got {len(batch_res_docs)}'
                        )
                elif batch_res_docs is None:
                    if not len(docs_inner_batch) == input_len_before_call:
                        raise ValueError(
                            f'Dynamic Batching requires input size to equal output size. Expected output size {input_len_before_call}, but got {len(docs_inner_batch)}'
                        )
                else:
                    array_name = (
                        'DocumentArray' if not docarray_v2 else 'DocList'
                    )
                    raise TypeError(
                        f'The return type must be {array_name} / `None` when using dynamic batching, '
                        f'but getting {batch_res_docs!r}'
                    )
            except Exception as exc:
                # All the requests containing docs in this Exception should be raising it
                for request_full in requests_completed_in_batch[
                                    involved_requests_min_indx: involved_requests_max_indx + 1
                                    ]:
                    await request_full.put(exc)
            else:
                # We need to attribute the docs to their requests
                non_assigned_to_response_docs.extend(
                    batch_res_docs or docs_inner_batch
                )
                non_assigned_to_response_request_idxs.extend(req_idxs)
                num_assigned_docs = await _assign_results(
                    non_assigned_to_response_docs,
                    non_assigned_to_response_request_idxs,
                    sum_from_previous_first_req_idx,
                    requests_lens_in_batch,
                    requests_in_batch,
                    requests_completed_in_batch,
                )

                sum_from_previous_first_req_idx = (
                        len(non_assigned_to_response_docs) - num_assigned_docs
                )
                non_assigned_to_response_docs = non_assigned_to_response_docs[
                                                num_assigned_docs:
                                                ]
                non_assigned_to_response_request_idxs = (
                    non_assigned_to_response_request_idxs[num_assigned_docs:]
                )
        if len(non_assigned_to_response_request_idxs) > 0:
            _ = await _assign_results(
                non_assigned_to_response_docs,
                non_assigned_to_response_request_idxs,
                sum_from_previous_first_req_idx,
                requests_lens_in_batch,
                requests_in_batch,
                requests_completed_in_batch,
            )


    async def close(self):
        """Closes the batch queue by flushing pending requests."""
        if not self._is_closed:
            # debug print amount of requests to be processed.
            self._flush_trigger.set()
            if self._flush_task:
                await self._flush_task
            self._cancel_timer_if_pending()
            self._is_closed = True



from jina.clients.base.http import HTTPBaseClient
from jina.clients.mixin import (
    AsyncHealthCheckMixin,
    AsyncMutateMixin,
    AsyncPostMixin,
    AsyncProfileMixin,
    HealthCheckMixin,
    MutateMixin,
    PostMixin,
    ProfileMixin,
)
import asyncio


class HTTPClient(
    HTTPBaseClient, PostMixin, ProfileMixin, MutateMixin, HealthCheckMixin
):
    """A client connecting to a Gateway using gRPC protocol.

    Instantiate this class through the :meth:`jina.Client` convenience method.

    EXAMPLE USAGE

    .. code-block:: python

        from jina import Client
        from docarray import Document

        # select host address to connect to
        c = Client(
            protocol='http', asyncio=False, host='http://my.awesome.flow:1234'
        )  # returns HTTPClient instance
        c.post(on='/index', inputs=Document(text='hello!'))

    """


class AsyncHTTPClient(
    HTTPBaseClient,
    AsyncPostMixin,
    AsyncMutateMixin,
    AsyncProfileMixin,
    AsyncHealthCheckMixin,
):
    """
    Asynchronous client connecting to a Gateway using HTTP protocol.

    Instantiate this class through the :meth:`jina.Client` convenience method.

    Unlike :class:`HTTPClient`, here :meth:`post` is a coroutine (i.e. declared with the async/await syntax),
    simply calling them will not schedule them to be executed.

    To actually run a coroutine, user need to put them in an event loop, e.g. via ``asyncio.run()``,
    ``asyncio.create_task()``.

    :class:`AsyncHTTPClient` can be very useful in
    the integration settings, where Jina/Flow/Client is NOT the main logic, but rather served as a part of other program.
    In this case, users often do not want to let Jina control the ``asyncio.eventloop``. On contrary, :class:`Client`
    is controlling and wrapping the event loop internally, making the Client looks synchronous from outside.

    EXAMPLE USAGE

    .. code-block:: python

        from jina import Client
        from docarray import Document

        # async inputs for the client
        async def async_inputs():
            for _ in range(10):
                yield Document()
                await asyncio.sleep(0.1)


        # select host address to connect to
        c = Client(
            protocol='http', asyncio=True, host='http://my.awesome.flow:1234'
        )  # returns AsyncHTTPClient instance

        async for resp in client.post(on='/index', async_inputs, request_size=1):
            print(resp)

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lock = asyncio.Lock()
        self.reuse_session = self.args.reuse_session



import argparse
import asyncio
import json
import os
from collections import defaultdict
from typing import TYPE_CHECKING, AsyncIterator, Dict, List, Optional, Tuple, Any

import grpc

from jina.enums import PollingType
from jina.constants import __default_endpoint__
from jina.excepts import InternalNetworkError
from jina.helper import get_full_version
from jina.proto import jina_pb2
from jina.serve.networking import GrpcConnectionPool
from jina.serve.runtimes.monitoring import MonitoringRequestMixin
from jina.serve.runtimes.worker.request_handling import WorkerRequestHandler
from jina.types.request.data import DataRequest, Response
from jina._docarray import docarray_v2

if docarray_v2:
    from jina.serve.runtimes.helper import _create_pydantic_model_from_schema
    from docarray import DocList
    from docarray.base_doc.any_doc import AnyDoc


if TYPE_CHECKING:  # pragma: no cover
    from prometheus_client import CollectorRegistry

    from jina.logging.logger import JinaLogger
    from jina.types.request import Request
    from jina.types.request.data import DataRequest


class HeaderRequestHandler(MonitoringRequestMixin):
    """
    Class that handles the requests arriving to the head and the results extracted from the requests future.

    :param metrics_registry: optional metrics registry for prometheus. Used if we need to expose metrics from the executor or from the data request handler
    :param runtime_name: optional runtime_name that will be registered during monitoring
    """

    DEFAULT_POLLING = PollingType.ANY

    def __init__(
        self,
        args: 'argparse.Namespace',
        logger: 'JinaLogger',
        metrics_registry: Optional['CollectorRegistry'] = None,
        meter=None,
        runtime_name: Optional[str] = None,
        aio_tracing_client_interceptors=None,
        tracing_client_interceptor=None,
        **kwargs,
    ):
        if args.name is None:
            args.name = ''
        self.logger = logger
        self.args = args
        self.meter = meter
        self.metrics_registry = metrics_registry
        self.name = args.name
        self._deployment_name = os.getenv('JINA_DEPLOYMENT_NAME', 'worker')
        self.aio_tracing_client_interceptors = aio_tracing_client_interceptors
        self.tracing_client_interceptor = tracing_client_interceptor
        self.connection_pool = GrpcConnectionPool(
            runtime_name=self.name,
            logger=self.logger,
            compression=args.compression,
            metrics_registry=self.metrics_registry,
            meter=self.meter,
            aio_tracing_client_interceptors=self.aio_tracing_client_interceptors,
            tracing_client_interceptor=self.tracing_client_interceptor,
            channel_options=self.args.grpc_channel_options,
        )
        self._retries = self.args.retries

        polling = getattr(args, 'polling', self.DEFAULT_POLLING.name)
        try:
            # try loading the polling args as json
            endpoint_polling = json.loads(polling)
            # '*' is used a wildcard and will match all endpoints, except /index, /search and explicitly defined endpoins
            default_polling = (
                PollingType.from_string(endpoint_polling['*'])
                if '*' in endpoint_polling
                else self.DEFAULT_POLLING
            )
            self._polling = self._default_polling_dict(default_polling)
            for endpoint in endpoint_polling:
                self._polling[endpoint] = PollingType(
                    endpoint_polling[endpoint]
                    if type(endpoint_polling[endpoint]) == int
                    else PollingType.from_string(endpoint_polling[endpoint])
                )
        except (ValueError, TypeError):
            # polling args is not a valid json, try interpreting as a polling enum type
            default_polling = (
                polling
                if type(polling) == PollingType
                else PollingType.from_string(polling)
            )
            self._polling = self._default_polling_dict(default_polling)

        if hasattr(args, 'connection_list') and args.connection_list:
            connection_list = json.loads(args.connection_list)
            for shard_id in connection_list:
                shard_connections = connection_list[shard_id]
                if isinstance(shard_connections, str):
                    self.connection_pool.add_connection(
                        deployment=self._deployment_name,
                        address=shard_connections,
                        shard_id=int(shard_id),
                    )
                else:
                    for connection in shard_connections:
                        self.connection_pool.add_connection(
                            deployment=self._deployment_name,
                            address=connection,
                            shard_id=int(shard_id),
                        )

        self.uses_before_address = args.uses_before_address
        self.timeout_send = args.timeout_send
        if self.timeout_send:
            self.timeout_send /= 1e3  # convert ms to seconds

        if self.uses_before_address:
            self.connection_pool.add_connection(
                deployment='uses_before', address=self.uses_before_address
            )
        self.uses_after_address = args.uses_after_address
        if self.uses_after_address:
            self.connection_pool.add_connection(
                deployment='uses_after', address=self.uses_after_address
            )
        self._reduce = not args.no_reduce
        super().__init__(metrics_registry, meter, runtime_name)
        self.logger = logger
        self._executor_endpoint_mapping = None
        self._gathering_endpoints = False
        self.runtime_name = runtime_name
        self._pydantic_models_by_endpoint = None
        self.endpoints_discovery_stop_event = asyncio.Event()
        self.endpoints_discovery_task = None
        if docarray_v2:
            self.endpoints_discovery_task = asyncio.create_task(
                self._get_endpoints_from_workers(
                    connection_pool=self.connection_pool,
                    name=self._deployment_name,
                    retries=self._retries,
                    stop_event=self.endpoints_discovery_stop_event,
                )
            )

    def _default_polling_dict(self, default_polling):
        return defaultdict(
            lambda: default_polling,
            {'/search': PollingType.ALL, '/index': PollingType.ANY},
        )

    async def _gather_worker_tasks(
        self,
        requests,
        connection_pool,
        deployment_name,
        polling_type,
        timeout_send,
        retries,
    ):
        worker_send_tasks = connection_pool.send_requests(
            requests=requests,
            deployment=deployment_name,
            polling_type=polling_type,
            timeout=timeout_send,
            retries=retries,
        )

        all_worker_results = await asyncio.gather(*worker_send_tasks)
        worker_results = list(
            filter(lambda x: isinstance(x, Tuple), all_worker_results)
        )
        exceptions = list(
            filter(
                lambda x: issubclass(type(x), BaseException),
                all_worker_results,
            )
        )
        total_shards = len(worker_send_tasks)
        failed_shards = len(exceptions)
        if failed_shards:
            self.logger.warning(f'{failed_shards} shards out of {total_shards} failed.')

        return worker_results, exceptions, total_shards, failed_shards

    @staticmethod
    def _merge_metadata(
        metadata,
        uses_after_metadata,
        uses_before_metadata,
        total_shards,
        failed_shards,
    ):
        merged_metadata = {}
        if uses_before_metadata:
            for key, value in uses_before_metadata:
                merged_metadata[key] = value
        for meta in metadata:
            for key, value in meta:
                merged_metadata[key] = value
        if uses_after_metadata:
            for key, value in uses_after_metadata:
                merged_metadata[key] = value

        merged_metadata['total_shards'] = str(total_shards)
        merged_metadata['failed_shards'] = str(failed_shards)
        return merged_metadata

    async def _handle_data_request(
        self,
        requests,
        connection_pool,
        uses_before_address,
        uses_after_address,
        timeout_send,
        retries,
        reduce,
        polling_type,
        deployment_name,
        endpoint,
    ) -> Tuple['DataRequest', Dict]:
        for req in requests:
            if docarray_v2:
                req.document_array_cls = DocList[AnyDoc]
            self._update_start_request_metrics(req)

        if self._pydantic_models_by_endpoint is None and docarray_v2:
            try:
                await self.endpoints_discovery_task
            except:
                raise

        WorkerRequestHandler.merge_routes(requests)

        uses_before_metadata = None
        if uses_before_address:
            result = await connection_pool.send_requests_once(
                requests,
                deployment='uses_before',
                timeout=timeout_send,
                retries=retries,
            )
            if issubclass(type(result), BaseException):
                raise result
            else:
                response, uses_before_metadata = result
                requests = [response]

        (
            worker_results,
            exceptions,
            total_shards,
            failed_shards,
        ) = await self._gather_worker_tasks(
            requests=requests,
            deployment_name=deployment_name,
            timeout_send=timeout_send,
            connection_pool=connection_pool,
            polling_type=polling_type,
            retries=retries,
        )
        if len(worker_results) == 0:
            if exceptions:
                # raise the underlying error first
                self._update_end_failed_requests_metrics()
                raise exceptions[0]
            raise RuntimeError(
                f'Head {self.runtime_name} did not receive a response when sending message to worker pods'
            )

        worker_results, metadata = zip(*worker_results)

        response_request = worker_results[0]
        found = False
        if docarray_v2:
            check_endpoint = endpoint
            if endpoint not in self._pydantic_models_by_endpoint:
                check_endpoint = __default_endpoint__
            model = self._pydantic_models_by_endpoint[check_endpoint]['output']
        for i, worker_result in enumerate(worker_results):
            if docarray_v2:
                worker_result.document_array_cls = DocList[model]
            if not found and worker_result.status.code == jina_pb2.StatusProto.SUCCESS:
                response_request = worker_result
                found = True

        uses_after_metadata = None
        if uses_after_address:
            result = await connection_pool.send_requests_once(
                worker_results,
                deployment='uses_after',
                timeout=timeout_send,
                retries=retries,
            )
            if issubclass(type(result), BaseException):
                self._update_end_failed_requests_metrics()
                raise result
            else:
                response_request, uses_after_metadata = result
        elif len(worker_results) > 1 and reduce:
            response_request = WorkerRequestHandler.reduce_requests(worker_results)
        elif len(worker_results) > 1 and not reduce:
            # worker returned multiple responses, but the head is configured to skip reduction
            # just concatenate the docs in this case
            response_request.data.docs = WorkerRequestHandler.get_docs_from_request(
                requests
            )

        merged_metadata = self._merge_metadata(
            metadata,
            uses_after_metadata,
            uses_before_metadata,
            total_shards,
            failed_shards,
        )

        self._update_end_request_metrics(response_request)
        return response_request, merged_metadata

    def _get_endpoints_from_workers(
        self, connection_pool: GrpcConnectionPool, name: str, retries: int, stop_event
    ):
        from google.protobuf import json_format
        from docarray.documents.legacy import LegacyDocument

        legacy_doc_schema = LegacyDocument.schema()

        async def task():
            self.logger.debug(
                f'starting get endpoints from workers task for deployment {name}'
            )
            while not stop_event.is_set():
                try:
                    endpoints = await connection_pool.send_discover_endpoint(
                        name, retries=retries, shard_id=0, head=False
                    )
                    if endpoints is not None:
                        endp, _ = endpoints
                        schemas = json_format.MessageToDict(endp.schemas)
                        self._pydantic_models_by_endpoint = {}
                        models_created_by_name = {}
                        for endpoint, inner_dict in schemas.items():
                            input_model_name = inner_dict['input']['name']
                            input_model_schema = inner_dict['input']['model']
                            output_model_name = inner_dict['output']['name']
                            output_model_schema = inner_dict['output']['model']

                            if input_model_schema == legacy_doc_schema:
                                models_created_by_name[input_model_name] = (
                                    LegacyDocument
                                )
                            elif input_model_name not in models_created_by_name:
                                input_model = _create_pydantic_model_from_schema(
                                    input_model_schema, input_model_name, {}
                                )
                                models_created_by_name[input_model_name] = input_model

                            if output_model_name == legacy_doc_schema:
                                models_created_by_name[output_model_name] = (
                                    LegacyDocument
                                )
                            elif output_model_name not in models_created_by_name:
                                output_model = _create_pydantic_model_from_schema(
                                    output_model_schema, output_model_name, {}
                                )
                                models_created_by_name[output_model_name] = output_model

                            self._pydantic_models_by_endpoint[endpoint] = {
                                'input': models_created_by_name[input_model_name],
                                'output': models_created_by_name[output_model_name],
                            }
                        stop_event.set()
                        return
                    else:
                        await asyncio.sleep(0.1)
                except Exception as exc:
                    self.logger.debug(
                        f'Exception raised from sending discover endpoint {exc}'
                    )
                    await asyncio.sleep(0.1)

        return task()

    def cancel_endpoint_discovery_from_workers_task(self):
        """Cancel endpoint_discovery_from_worker task if exists and is not completed. Cancellation is required if the Flow is being terminated before the
        task is successful or hasn't reached the max timeout.
        """
        if self.endpoints_discovery_task:
            try:
                if not self.endpoints_discovery_task.done():
                    self.logger.debug(f'Cancelling endpoint discovery task.')
                    self.endpoints_discovery_stop_event.set()  # this event is useless if simply cancel
                    self.endpoints_discovery_task.cancel()
            except Exception as ex:
                self.logger.debug(
                    f'exception during endpoint discovery task cancellation: {ex}'
                )
                pass

    async def close(self):
        """Close the data request handler, by closing the executor and the batch queues."""
        self.logger.debug(f'Closing Request Handler')
        self.cancel_endpoint_discovery_from_workers_task()
        await self.connection_pool.close()
        self.logger.debug(f'Request Handler closed')

    async def process_single_data(self, request: DataRequest, context) -> DataRequest:
        """
        Process the received requests and return the result as a new request

        :param request: the data request to process
        :param context: grpc context
        :returns: the response request
        """
        return await self.process_data([request], context)

    def _handle_internalnetworkerror(self, err, context, response):
        err_code = err.code()
        if err_code == grpc.StatusCode.UNAVAILABLE:
            context.set_details(
                f'|Head: Failed to connect to worker (Executor) pod at address {err.dest_addr}. It may be down.'
            )
        elif err_code == grpc.StatusCode.DEADLINE_EXCEEDED:
            context.set_details(
                f'|Head: Connection to worker (Executor) pod at address {err.dest_addr} could be established, but timed out.'
            )
        elif err_code == grpc.StatusCode.NOT_FOUND:
            context.set_details(
                f'|Head: Connection to worker (Executor) pod at address {err.dest_addr} could be established, but resource was not found.'
            )
        context.set_code(err.code())
        self.logger.error(f'Error while getting responses from Pods: {err.details()}')
        if err.request_id:
            response.header.request_id = err.request_id
        return response

    async def process_data(self, requests: List[DataRequest], context) -> DataRequest:
        """
        Process the received data request and return the result as a new request

        :param requests: the data requests to process
        :param context: grpc context
        :returns: the response request
        """
        try:
            endpoint = dict(context.invocation_metadata()).get('endpoint')
            self.logger.debug(f'recv {len(requests)} DataRequest(s)')
            response, metadata = await self._handle_data_request(
                requests=requests,
                connection_pool=self.connection_pool,
                uses_before_address=self.uses_before_address,
                uses_after_address=self.uses_after_address,
                retries=self._retries,
                reduce=self._reduce,
                timeout_send=self.timeout_send,
                polling_type=self._polling[endpoint],
                deployment_name=self._deployment_name,
                endpoint=endpoint,
            )
            context.set_trailing_metadata(metadata.items())
            return response
        except (
            InternalNetworkError
        ) as err:  # can't connect, Flow broken, interrupt the streaming through gRPC error mechanism
            return self._handle_internalnetworkerror(
                err=err, context=context, response=Response()
            )
        except (
            RuntimeError,
            Exception,
        ) as ex:  # some other error, keep streaming going just add error info
            self.logger.error(
                (
                    f'{ex!r}'
                    + f'\n add "--quiet-error" to suppress the exception details'
                    if not self.args.quiet_error
                    else ''
                ),
                exc_info=not self.args.quiet_error,
            )
            requests[0].add_exception(ex, executor=None)
            context.set_trailing_metadata((('is-error', 'true'),))
            return requests[0]

    async def endpoint_discovery(self, empty, context) -> jina_pb2.EndpointsProto:
        """
        Uses the connection pool to send a discover endpoint call to the workers

        :param empty: The service expects an empty protobuf message
        :param context: grpc context
        :returns: the response request
        """
        self.logger.debug('got an endpoint discovery request')
        response = jina_pb2.EndpointsProto()
        try:
            if self.uses_before_address:
                (
                    uses_before_response,
                    _,
                ) = await self.connection_pool.send_discover_endpoint(
                    deployment='uses_before', head=False
                )
                response.endpoints.extend(uses_before_response.endpoints)
            if self.uses_after_address:
                (
                    uses_after_response,
                    _,
                ) = await self.connection_pool.send_discover_endpoint(
                    deployment='uses_after', head=False
                )
                response.endpoints.extend(uses_after_response.endpoints)

            worker_response, _ = await self.connection_pool.send_discover_endpoint(
                deployment=self._deployment_name, head=False
            )
            response.endpoints.extend(worker_response.endpoints)
            response.schemas.update(worker_response.schemas)
        except (
            InternalNetworkError
        ) as err:  # can't connect, Flow broken, interrupt the streaming through gRPC error mechanism
            return self._handle_internalnetworkerror(
                err=err, context=context, response=response
            )

        return response

    async def _status(self, empty, context) -> jina_pb2.JinaInfoProto:
        """
        Process the the call requested and return the JinaInfo of the Runtime

        :param empty: The service expects an empty protobuf message
        :param context: grpc context
        :returns: the response request
        """
        self.logger.debug('recv _status request')
        infoProto = jina_pb2.JinaInfoProto()
        version, env_info = get_full_version()
        for k, v in version.items():
            infoProto.jina[k] = str(v)
        for k, v in env_info.items():
            infoProto.envs[k] = str(v)
        return infoProto

    async def stream(
        self, request_iterator, context=None, *args, **kwargs
    ) -> AsyncIterator['Request']:
        """
        stream requests from client iterator and stream responses back.

        :param request_iterator: iterator of requests
        :param context: context of the grpc call
        :param args: positional arguments
        :param kwargs: keyword arguments
        :yield: responses to the request
        """
        async for request in request_iterator:
            yield await self.process_data([request], context)

    Call = stream



import asyncio
from typing import AsyncIterator, Iterator, Optional, Union

from jina.helper import get_or_reuse_loop


class _RequestsCounter:
    """Class used to wrap a count integer so that it can be updated inside methods.

    .. code-block:: python

        def count_increment(i: int, rc: _RequestsCounter):
            i += 1
            rc.count += 1


        c_int = 0
        c_rc = _RequestsCounter()
        count_increment(c_int, c_rc)

        assert c_int == 0
        assert c_rc.count == 1
    """

    count = 0


class AsyncRequestsIterator:
    """Iterator to allow async iteration of blocking/non-blocking iterator from the Client"""

    def __init__(
        self,
        iterator: Union[Iterator, AsyncIterator],
        request_counter: Optional[_RequestsCounter] = None,
        prefetch: int = 0,
        iterate_sync_in_thread: bool = True,
    ) -> None:
        """Async request iterator

        :param iterator: request iterator
        :param request_counter: counter of the numbers of request being handled at a given moment
        :param prefetch: The max amount of requests to be handled at a given moment (0 disables feature)
        :param iterate_sync_in_thread: if True, blocking iterators will call __next__ in a Thread.
        """
        self.iterator = iterator
        self._request_counter = request_counter
        self._prefetch = prefetch
        self._iterate_sync_in_thread = iterate_sync_in_thread

    def iterator__next__(self):
        """
        Executed inside a `ThreadPoolExecutor` via `loop.run_in_executor` to avoid following exception.
        "StopIteration interacts badly with generators and cannot be raised into a Future"

        :return: next request or None
        """
        try:
            return self.iterator.__next__()
        except StopIteration:
            return None

    def __aiter__(self):
        return self

    async def __anext__(self):
        if isinstance(self.iterator, Iterator):
            """
            An `Iterator` indicates "blocking" code, which might block all tasks in the event loop.
            Hence we iterate in the default executor provided by asyncio.
            """

            if not self._iterate_sync_in_thread:

                async def _get_next():
                    try:
                        req = self.iterator.__next__()
                    except StopIteration:
                        req = None
                    return req

                request = await asyncio.create_task(_get_next())
            else:
                request = await get_or_reuse_loop().run_in_executor(
                    None, self.iterator__next__
                )

            """
            `iterator.__next__` can be executed directly and that'd raise `StopIteration` in the executor,
            which raises the following exception while chaining states in futures.
            "StopIteration interacts badly with generators and cannot be raised into a Future"
            To avoid that, we handle the raise by a `return None`
            """
            if request is None:
                raise StopAsyncIteration
        elif isinstance(self.iterator, AsyncIterator):
            # we assume that `AsyncIterator` doesn't block the event loop
            request = await self.iterator.__anext__()

        if self._prefetch > 0:
            while self._request_counter.count >= self._prefetch:
                await asyncio.sleep(0)
        return request
